@@ -4,6 +4,7 @@
 
 import base64
 import os
+import pathlib
 import re
 import socket
 
@@ -12,7 +13,9 @@ import pytest
 import spnego
 import spnego._ntlm as ntlm
 import spnego.channel_bindings
+import spnego.iov
 from spnego._credential import CredentialCache
+from spnego._ntlm_raw.crypto import md4
 from spnego._ntlm_raw.messages import (
     Authenticate,
     AvId,
@@ -27,7 +30,6 @@ from spnego._text import to_bytes, to_text
 from spnego.exceptions import (
     BadBindingsError,
     BadMICError,
-    FeatureMissingError,
     InvalidTokenError,
     NoContextError,
     OperationNotAvailableError,
@@ -166,16 +168,24 @@ def test_invalid_lm_compat_level(level, monkeypatch):
         ntlm.NTLMProxy("user", "pass")
 
 
-@pytest.mark.parametrize("usage", ["initiate", "accept"])
-def test_context_no_store(usage):
+def test_context_no_store_initiate():
     with pytest.raises(
-        OperationNotAvailableError, match="Retrieving NTLM store without NTLM_USER_FILE set to a " "filepath"
+        OperationNotAvailableError,
+        match="No username or password was specified and the credential cache did not exist or contained no credentials",
     ):
-        ntlm.NTLMProxy(CredentialCache(), usage=usage)
+        ntlm.NTLMProxy(CredentialCache(), usage="initiate")
+
+
+def test_context_no_store_accept():
+    with pytest.raises(
+        OperationNotAvailableError,
+        match="NTLM acceptor requires NTLM credential cache to be provided through the env var NTLM_USER_FILE set to a filepath",
+    ):
+        ntlm.NTLMProxy(CredentialCache(), usage="accept")
 
 
 def test_iov_available():
-    assert ntlm.NTLMProxy.iov_available() is False
+    assert ntlm.NTLMProxy.iov_available() is True
 
 
 def test_ntlm_invalid_usage():
@@ -188,13 +198,10 @@ def test_ntlm_invalid_protocol():
         ntlm.NTLMProxy("user", "pass", protocol="fake")
 
 
-def test_ntlm_iov_not_available():
-    expected = (
-        "The system is missing the GSSAPI IOV extension headers or NTLM or CredSSP is being requested, "
-        "cannot utilize wrap_iov and unwrap_iov"
-    )
-    with pytest.raises(FeatureMissingError, match=re.escape(expected)):
-        ntlm.NTLMProxy("user", "pass", options=spnego.NegotiateOptions.wrapping_iov)
+def test_ntlm_query_message_sizes_fail():
+    n = ntlm.NTLMProxy("user", "pass")
+    with pytest.raises(NoContextError, match="Cannot get message sizes until context has been established"):
+        n.query_message_sizes()
 
 
 def test_ntlm_wrap_qop_invalid():
@@ -235,18 +242,164 @@ def test_ntlm_unwrap_winrm_no_context():
         n.unwrap_winrm(b"header", b"data")
 
 
-def test_ntlm_wrap_iov_fail():
+def test_ntlm_wrap_iov_with_qop_fail():
     n = ntlm.NTLMProxy("user", "pass")
-    with pytest.raises(OperationNotAvailableError, match="NTLM does not offer IOV wrapping"):
+    with pytest.raises(UnsupportedQop, match="Unsupported QoP value 1 specified for NTLM"):
+        n.wrap_iov([], qop=1)
+
+
+def test_ntlm_wrap_iov_no_sign_or_seal():
+    n = ntlm.NTLMProxy("user", "pass")
+    with pytest.raises(OperationNotAvailableError, match="NTLM wrap without integrity or confidentiality"):
         n.wrap_iov([])
 
 
-def test_ntlm_unwrap_iov_fail():
+def test_ntlm_wrap_iov_no_context():
     n = ntlm.NTLMProxy("user", "pass")
-    with pytest.raises(OperationNotAvailableError, match="NTLM does not offer IOV wrapping"):
+    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
+    with pytest.raises(NoContextError, match="Cannot wrap until context has been established"):
+        n.wrap_iov([])
+
+
+def test_ntlm_wrap_iov_no_header():
+    n = ntlm.NTLMProxy("user", "pass")
+    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
+    n._sign_key_out = n._handle_out = 1  # type: ignore
+
+    with pytest.raises(InvalidTokenError, match="wrap_iov no IOV header buffer present"):
+        n.wrap_iov([b""])
+
+
+def test_ntlm_wrap_iov_no_data():
+    n = ntlm.NTLMProxy("user", "pass")
+    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
+    n._sign_key_out = n._handle_out = 1  # type: ignore
+
+    with pytest.raises(InvalidTokenError, match="wrap_iov no IOV data buffer present"):
+        n.wrap_iov([spnego.iov.BufferType.header])
+
+
+def test_ntlm_wrap_iov_multiple_header():
+    n = ntlm.NTLMProxy("user", "pass")
+    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
+    n._sign_key_out = n._handle_out = 1  # type: ignore
+
+    with pytest.raises(InvalidTokenError, match="wrap_iov must only be used with 1 header IOV buffer"):
+        n.wrap_iov([spnego.iov.BufferType.header, b"", spnego.iov.BufferType.header])
+
+
+def test_ntlm_wrap_iov_multiple_data():
+    n = ntlm.NTLMProxy("user", "pass")
+    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
+    n._sign_key_out = n._handle_out = 1  # type: ignore
+
+    with pytest.raises(InvalidTokenError, match="wrap_iov must only be used with 1 data IOV buffer"):
+        n.wrap_iov([spnego.iov.BufferType.header, b"", b""])
+
+
+def test_ntlm_wrap_iov_invalid_type():
+    n = ntlm.NTLMProxy("user", "pass")
+    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
+    n._sign_key_out = n._handle_out = 1  # type: ignore
+
+    with pytest.raises(InvalidTokenError, match="wrap_iov unsupported IOV buffer type padding"):
+        n.wrap_iov([spnego.iov.BufferType.header, b"", spnego.iov.BufferType.padding])
+
+
+def test_ntlm_wrap_iov_data_not_bytes():
+    n = ntlm.NTLMProxy("user", "pass")
+    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
+    n._sign_key_out = n._handle_out = 1  # type: ignore
+
+    with pytest.raises(InvalidTokenError, match="wrap_iov IOV data buffer at \\[1\\] must be bytes"):
+        n.wrap_iov([spnego.iov.BufferType.header, (spnego.iov.BufferType.data, 1)])
+
+
+def test_ntlm_wrap_iov_signonly_not_bytes():
+    n = ntlm.NTLMProxy("user", "pass")
+    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
+    n._sign_key_out = n._handle_out = 1  # type: ignore
+
+    with pytest.raises(InvalidTokenError, match="wrap_iov IOV sign_only buffer at \\[1\\] must be bytes"):
+        n.wrap_iov([spnego.iov.BufferType.header, (spnego.iov.BufferType.sign_only, 1)])
+
+
+def test_ntlm_unwrap_iov_no_sign_or_seal():
+    n = ntlm.NTLMProxy("user", "pass")
+    with pytest.raises(OperationNotAvailableError, match="NTLM unwrap without integrity or confidentiality"):
         n.unwrap_iov([])
 
 
+def test_ntlm_unwrap_iov_no_context():
+    n = ntlm.NTLMProxy("user", "pass")
+    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
+    with pytest.raises(NoContextError, match="Cannot unwrap until context has been established"):
+        n.unwrap_iov([])
+
+
+def test_ntlm_unwrap_iov_no_header():
+    n = ntlm.NTLMProxy("user", "pass")
+    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
+    n._sign_key_in = n._handle_in = 1  # type: ignore
+
+    with pytest.raises(InvalidTokenError, match="unwrap_iov no IOV header buffer present"):
+        n.unwrap_iov([b""])
+
+
+def test_ntlm_unwrap_iov_no_data():
+    n = ntlm.NTLMProxy("user", "pass")
+    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
+    n._sign_key_in = n._handle_in = 1  # type: ignore
+
+    with pytest.raises(InvalidTokenError, match="unwrap_iov no IOV data buffer present"):
+        n.unwrap_iov([(spnego.iov.BufferType.header, b"")])
+
+
+def test_ntlm_unwrap_iov_multiple_header():
+    n = ntlm.NTLMProxy("user", "pass")
+    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
+    n._sign_key_in = n._handle_in = 1  # type: ignore
+
+    with pytest.raises(InvalidTokenError, match="unwrap_iov must only be used with 1 header IOV buffer"):
+        n.unwrap_iov([(spnego.iov.BufferType.header, b""), b"", (spnego.iov.BufferType.header, b"")])
+
+
+def test_ntlm_unwrap_iov_multiple_data():
+    n = ntlm.NTLMProxy("user", "pass")
+    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
+    n._sign_key_in = n._handle_in = 1  # type: ignore
+
+    with pytest.raises(InvalidTokenError, match="unwrap_iov must only be used with 1 data IOV buffer"):
+        n.unwrap_iov([(spnego.iov.BufferType.header, b""), b"", b""])
+
+
+def test_ntlm_unwrap_iov_invalid_type():
+    n = ntlm.NTLMProxy("user", "pass")
+    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
+    n._sign_key_in = n._handle_in = 1  # type: ignore
+
+    with pytest.raises(InvalidTokenError, match="unwrap_iov unsupported IOV buffer type padding"):
+        n.unwrap_iov([(spnego.iov.BufferType.header, b""), b"", (spnego.iov.BufferType.padding, b"")])
+
+
+def test_ntlm_unwrap_iov_data_not_bytes():
+    n = ntlm.NTLMProxy("user", "pass")
+    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
+    n._sign_key_in = n._handle_in = 1  # type: ignore
+
+    with pytest.raises(InvalidTokenError, match="unwrap_iov IOV data buffer at \\[1\\] must be bytes"):
+        n.unwrap_iov([(spnego.iov.BufferType.header, b""), (spnego.iov.BufferType.data, 1)])
+
+
+def test_ntlm_unwrap_iov_data_readonly_not_bytes():
+    n = ntlm.NTLMProxy("user", "pass")
+    n._context_attr = spnego.ContextReq.confidentiality | spnego.ContextReq.integrity
+    n._sign_key_in = n._handle_in = 1  # type: ignore
+
+    with pytest.raises(InvalidTokenError, match="unwrap_iov IOV data_readonly buffer at \\[1\\] must be bytes"):
+        n.unwrap_iov([(spnego.iov.BufferType.header, b""), (spnego.iov.BufferType.data_readonly, 1)])
+
+
 def test_ntlm_sign_qop_invalid():
     n = ntlm.NTLMProxy("user", "pass")
     with pytest.raises(UnsupportedQop, match="Unsupported QoP value 1 specified for NTLM"):
@@ -573,6 +726,72 @@ def test_ntlm_no_nt_v1_allowed(ntlm_cred, monkeypatch):
         s.step(auth)
 
 
+def test_ntlm_with_invalid_surrogate_pair_pass(
+    monkeypatch: pytest.MonkeyPatch,
+    tmp_path: pathlib.Path,
+) -> None:
+    username = "user"
+
+    # This is a password for a gMSA account used as a test. It is technically
+    # a UTF-16-LE string value but contains invalid surrogate pairs which is
+    # important for testing.
+    b_gmsa_password = (
+        b"\x91\x45\xC9\xD1\x1D\x74\xA9\xE2"
+        b"\x27\x5A\x4C\xBE\x13\xC1\xE2\xF4"
+        b"\x89\x94\x49\x5E\x01\x60\xDD\xBA"
+        b"\xE2\xD9\x12\x53\xF0\xEB\x96\x38"
+        b"\xB3\x8B\xD2\x17\xC9\xCC\x9B\xB6"
+        b"\xC7\xF0\xCC\x8F\xEB\x75\x03\x77"
+        b"\x30\xD3\xE2\x6C\xE6\x00\x04\x39"
+        b"\xF5\x5F\xD9\xA5\xD8\xEF\xB3\x9F"
+        b"\xDE\x4A\xB5\xC1\x51\xC0\x44\x3B"
+        b"\x66\xC6\xF4\x68\x8D\xE1\x78\xBE"
+        b"\x3D\x35\x34\xC7\x4A\x91\x6D\x7B"
+        b"\x3C\xE0\x6E\x1C\xE9\xA9\x96\x6E"
+        b"\xDA\x09\x6A\x39\x1A\x2E\x5F\xD2"
+        b"\x92\x86\x46\x3B\x8B\x9A\xD3\xCE"
+        b"\xED\x83\x03\x2A\x33\x0D\xBC\x06"
+        b"\x91\xC6\x0C\xB1\x69\x5D\x2D\x59"
+        b"\xE0\x66\x18\x99\x00\xD1\x5A\x55"
+        b"\x85\xA3\xA8\x23\x0E\xCC\x16\x08"
+        b"\xF9\xE1\x9D\xF7\x09\x24\x66\xB3"
+        b"\x56\x6D\xC8\x2B\x4D\x33\x7F\x1A"
+        b"\xED\x69\x24\x09\xB1\x0C\xD6\x51"
+        b"\xBB\x62\xD9\x82\xD4\xA6\x1D\x91"
+        b"\x6F\xC4\xB2\xB0\x45\x9A\x40\x5A"
+        b"\xEC\x81\x71\xA1\x48\xB3\x52\x37"
+        b"\x26\x72\x98\x01\x22\x31\xF2\xD8"
+        b"\xD4\x83\x7B\xF3\xCA\xD5\x81\x24"
+        b"\xDC\xA9\xC2\xBF\x6D\x8E\x87\x7D"
+        b"\x24\x87\x49\x6C\x46\xE6\x67\x8B"
+        b"\x10\x69\x00\x04\xCA\x17\x4B\xC8"
+        b"\x04\x33\x69\x06\x61\x57\xB9\xC7"
+        b"\x3B\xFC\x0A\xCD\x35\xCE\x61\xB9"
+        b"\x87\x3B\xFF\x3A\x2D\x55\x67\xF6"
+    )
+
+    tmp_creds = tmp_path / "ntlm.cred"
+    with open(tmp_creds, mode="w") as fd:
+        nt_hash = base64.b16encode(md4(b_gmsa_password)).decode()
+        fd.write(f"{username}:1:00000000000000000000000000000000:{nt_hash}:[U]:LCT-1589398321")
+
+    monkeypatch.setenv("NTLM_USER_FILE", str(tmp_creds.absolute()))
+
+    c = spnego.client(
+        username,
+        b_gmsa_password.decode("utf-16-le", errors="surrogatepass"),
+        hostname=socket.gethostname(),
+        options=spnego.NegotiateOptions.use_ntlm,
+        protocol="ntlm",
+    )
+    s = spnego.server(options=spnego.NegotiateOptions.use_ntlm, protocol="ntlm")
+
+    s.step(c.step(s.step(c.step())))
+
+    assert c.complete
+    assert s.complete
+
+
 @pytest.mark.parametrize(
     "client_opt",
     [
@@ -637,3 +856,131 @@ def test_ntlm_anon_response(ntlm_cred):
 
     with pytest.raises(OperationNotAvailableError, match="Anonymous user authentication not implemented"):
         s.step(anon_auth)
+
+
+def test_ntlm_iov_wrapping(ntlm_cred):
+    c = spnego.client(
+        ntlm_cred[0],
+        ntlm_cred[1],
+        hostname=socket.gethostname(),
+        protocol="ntlm",
+        options=spnego.NegotiateOptions.use_ntlm,
+    )
+    s = spnego.server(protocol="ntlm")
+
+    s.step(c.step(s.step(c.step())))
+
+    res1 = c.wrap_iov(
+        [
+            (spnego.iov.BufferType.sign_only, b"sign 1"),
+            b"data",
+            (spnego.iov.BufferType.data_readonly, b"sign 2"),
+            spnego.iov.BufferType.header,
+        ]
+    )
+    assert isinstance(res1, spnego.IOVWrapResult)
+    assert res1.encrypted
+    assert len(res1.buffers) == 4
+    assert res1.buffers[0].data == b"sign 1"
+    assert res1.buffers[1].data != b"data"
+    assert res1.buffers[2].data == b"sign 2"
+    assert len(res1.buffers[3].data or b"") == 16
+
+    res2 = s.unwrap_iov(res1.buffers)
+    assert isinstance(res2, spnego.IOVUnwrapResult)
+    assert res2.encrypted
+    assert res2.qop == 0
+    assert len(res2.buffers) == 4
+    assert res2.buffers[0].data == b"sign 1"
+    assert res2.buffers[1].data == b"data"
+    assert res2.buffers[2].data == b"sign 2"
+    assert res2.buffers[3].data == res1.buffers[3].data
+
+    res3 = s.wrap_iov(
+        [
+            (spnego.iov.BufferType.sign_only, b"sign 1"),
+            b"data",
+            (spnego.iov.BufferType.data_readonly, b"sign 2"),
+            spnego.iov.BufferType.header,
+        ]
+    )
+    assert isinstance(res3, spnego.IOVWrapResult)
+    assert res3.encrypted
+    assert len(res3.buffers) == 4
+    assert res3.buffers[0].data == b"sign 1"
+    assert res3.buffers[1].data != b"data"
+    assert res3.buffers[2].data == b"sign 2"
+    assert len(res3.buffers[3].data or b"") == 16
+
+    res4 = c.unwrap_iov(res3.buffers)
+    assert isinstance(res4, spnego.IOVUnwrapResult)
+    assert res4.encrypted
+    assert res4.qop == 0
+    assert len(res4.buffers) == 4
+    assert res4.buffers[0].data == b"sign 1"
+    assert res4.buffers[1].data == b"data"
+    assert res4.buffers[2].data == b"sign 2"
+    assert res4.buffers[3].data == res3.buffers[3].data
+
+
+def test_ntlm_iov_unwrapping_as_stream(ntlm_cred):
+    c = spnego.client(
+        ntlm_cred[0],
+        ntlm_cred[1],
+        hostname=socket.gethostname(),
+        protocol="ntlm",
+        options=spnego.NegotiateOptions.use_ntlm,
+    )
+    s = spnego.server(protocol="ntlm")
+
+    s.step(c.step(s.step(c.step())))
+
+    res1 = c.wrap_iov(
+        [
+            b"data",
+            spnego.iov.BufferType.header,
+        ]
+    )
+    assert isinstance(res1, spnego.IOVWrapResult)
+    assert res1.encrypted
+    assert len(res1.buffers) == 2
+    assert res1.buffers[0].data != b"data"
+    assert len(res1.buffers[1].data or b"") == 16
+
+    msg = (res1.buffers[1].data or b"") + (res1.buffers[0].data or b"")
+    res2 = s.unwrap_iov(
+        [
+            (spnego.iov.BufferType.stream, msg),
+            spnego.iov.BufferType.data,
+        ]
+    )
+    assert isinstance(res2, spnego.IOVUnwrapResult)
+    assert res2.encrypted
+    assert res2.qop == 0
+    assert len(res2.buffers) == 2
+    assert res2.buffers[1].data == b"data"
+
+    res3 = s.wrap_iov(
+        [
+            spnego.iov.BufferType.header,
+            b"data",
+        ]
+    )
+    assert isinstance(res3, spnego.IOVWrapResult)
+    assert res3.encrypted
+    assert len(res3.buffers) == 2
+    assert len(res3.buffers[0].data or b"") == 16
+    assert res3.buffers[1].data != b"data"
+
+    msg = (res3.buffers[0].data or b"") + (res3.buffers[1].data or b"")
+    res4 = c.unwrap_iov(
+        [
+            (spnego.iov.BufferType.stream, msg),
+            spnego.iov.BufferType.data,
+        ]
+    )
+    assert isinstance(res4, spnego.IOVUnwrapResult)
+    assert res4.encrypted
+    assert res4.qop == 0
+    assert len(res4.buffers) == 2
+    assert res4.buffers[1].data == b"data"
